#!/usr/bin/perl

# Copyright 2017 Dmitry Suplatov
#  
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or of
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# Program: aln2pdbannotation.pl
#  
# Version: v.1.5 2017-11-24
#  
# Author:  Dmitry Suplatov d.a.suplatov(at)belozersky.msu.ru
# 
# Webpage: http://biokinet.belozersky.msu.ru/mustguseal
#  
# Ref:     Suplatov D., Kopylov K., et al. Mustguseal: Server for Multiple 
#          Structure-Guided Sequence Alignments of Protein Families, in preparation
#  
# Aim:     Calculate basic alignment statistics and annotate available PDB structures accordingly";
#  

my $VERBOSE=0;
my $SHANNON_MAX = 20*(1/20)*log(20);
my $SHANNON_MIN = 0;

use strict;
use warnings;
use v5.16;
use Cwd;
use File::Basename;
use File::Copy;

#Check if the script has been launched with exactly one command-line argument 
if (@ARGV!=3) {
    print "\n";
    print "This script will calculate basic alignment statistics and annotate available PDB structures accordingly\n";
    print "By Dmitry Suplatov d.a.suplatov\@belozersky.msu.ru\n";
    print "v. 1.5 2017-11-24\n";
    print "\n";
    print "Usage: $0 <input.fasta> <folder with pdbs> <output folder>\n";
    print "                    use <folder with pdbs>=null to disable pdb annotation\n";
    print "\n";
    exit 0;
}

my $MAFFT = `which mafft`;
chomp $MAFFT;
if (!-x $MAFFT) {
    print "Error: Mafft binary was not found or is not executable\n";
    exit;
}
if ($VERBOSE) {print "Info: Using Mafft executable from $MAFFT\n";}

my $PYMOL = `which pymol`;
chomp $PYMOL;
if (!-x $PYMOL) {
    print "Error: Pymol binary was not found or is not executable\n";
    exit;
}
if ($VERBOSE) {print "Info: Using Pymol executable from $PYMOL\n";}

#The first command line argument is the path to CMAT file
my $FASTA = $ARGV[0];
#The second command line argument is the path to the pdb file
my $PDB_FOLDER = $ARGV[1];
#Third argument - output folder
my $OUTPUT = $ARGV[2];

my $MASK = fileparse($FASTA, "fasta");

if ( ! -e $FASTA || -z $FASTA ) {
    print "<Font color=red>Error: File $FASTA does not exist or has zero size</Font>\n";
    exit 1;
}

if ( $PDB_FOLDER ne "null" && ! -d $PDB_FOLDER) {
    print "<Font color=red>Error: PDB folder $PDB_FOLDER does not exist</Font>\n";
    exit 1;
}

if (-d $OUTPUT) {
    print "<Font color=red>Error: Output folder already exists</Font>\n";
    exit 1;
}

mkdir $OUTPUT or die "<Font color=red>Error: Failed to create output folder $OUTPUT</Font>\n";

#Create the hash to store the collected data
my %hash = ();

#Open the CMAT file for reading
my $aln = readaln($FASTA);
my @seqn = keys %{$aln};

print "\n";
print "Info: Calculating alignment statistics\n";

my $ANNOTATION = "$OUTPUT/annotation.html";
my $NSEQS = "$OUTPUT/nseqs.php";

open (FA, ">$ANNOTATION") || die "<Font color=red>Error: Failed to open file $ANNOTATION for writing</Font>\n";
open (FS, ">$NSEQS") || die "<Font color=red>Error: Failed to open file $NSEQS for writing</Font>\n";

my %columns_aa = ();
my %columns_gaps = ();
my %seqlength = ();
my $nseqs = scalar @seqn;
my $ncolumns = 0;
foreach my $seqname (@seqn) {
    my $seq = $aln->{$seqname};
    
    if ($ncolumns != 0 && $ncolumns != $#{$seq}+1 ) {
        print "<Font color=red>Error: Inconsistent number of columns ($ncolumns and ".($#{$seq}+1).")\n";
        exit 1;
    }    
    $ncolumns = $#{$seq}+1;
    
    my $curseqlength=0;
    foreach my $i (0..$#{$seq}){
        if ($seq->[$i] eq "-") {
            $columns_gaps{$i}++;
        }
        else {
            $curseqlength++;
            $columns_aa{$i}{$seq->[$i]}++;
        }
    }
    $seqlength{$seqname} = $curseqlength;
}
#1. General statistics
print FS "<?php \$nseqs=$nseqs; ?>\n";
print FA "<B>General alignment statistics:</B><BR>\n";
print FA "<Font color=grey>Total number of</Font> proteins <Font color=grey>in the final alignment:</Font> $nseqs<BR>\n";
print FA "<Font color=grey>Total number of</Font> columns <Font color=grey>in the final alignment:</Font>  $ncolumns<BR>\n";
print FA "<BR>\n";

#1. Sequence length
my $length_average = 0;
my $length_min = -1;
my $length_max = -1;
foreach my $seqname (keys %seqlength) {
    $length_average += $seqlength{$seqname};
    if ($length_min == -1 || $length_min > $seqlength{$seqname}) {
        $length_min = $seqlength{$seqname};
    }
    if ($length_max == -1 || $length_max < $seqlength{$seqname}) {
        $length_max = $seqlength{$seqname};
    }
}
$length_average = sprintf("%.0f", $length_average / $nseqs);
print FA "<B>Protein length statistics:</B><BR>\n";
print FA "<table style=\"width:35%\">\n";

print FA "<tr>\n";
print FA "  <td width=\"55%\" align=\"left\">\n";
print FA "	<Font color=grey>Protein length</Font> average<Font color=grey>:</Font>\n";
print FA "  </td>\n";
print FA "  <td width=\"45%\" align=\"left\">\n";
print FA "	$length_average aa\n";
print FA "  </td>\n";
print FA "</tr>\n";


print FA "<tr>\n";
print FA "  <td width=\"55%\" align=\"left\">\n";
print FA "	<Font color=grey>Protein length</Font> minimum<Font color=grey>:</Font>\n";
print FA "  </td>\n";
print FA "  <td width=\"45%\" align=\"left\">\n";
print FA "	$length_min aa\n";
print FA "  </td>\n";
print FA "</tr>\n";


print FA "<tr>\n";
print FA "  <td width=\"55%\" align=\"left\">\n";
print FA "	<Font color=grey>Protein length</Font> maximum<Font color=grey>:</Font>\n";
print FA "  </td>\n";
print FA "  <td width=\"45%\" align=\"left\">\n";
print FA "	$length_max aa\n";
print FA "  </td>\n";
print FA "</tr>\n";

print FA "</table>\n";
print FA "<BR>\n";

#2. Calculate gaps statistics
print FA "<B>Alignment coverage statistics:</B><BR>\n";
print FA "<table style=\"width:70%\">\n";
foreach my $i (0..$ncolumns-1) {
    if (!$columns_gaps{$i}) {
        $columns_gaps{$i} = 0;
    }
    else {
        $columns_gaps{$i} = 100*$columns_gaps{$i}/$nseqs;
    }
}
my @gap_threshold = (0, 5, 30, 50);
foreach my $th (@gap_threshold) {
    my $counter = 0;
    foreach my $i (0..$ncolumns-1) {
        if ($columns_gaps{$i} <= $th) {$counter++;}
    }
    print FA "<tr>\n";
    print FA "  <td width=\"43%\" align=\"left\">\n";
    print FA "    <Font color=grey>Number of columns with at most</Font> ".sprintf("%3s", $th). "% <Font color=grey>of gaps:</Font>\n";
    print FA "  </td>\n";
    print FA "  <td width=\"10%\" align=\"center\">\n";
    print FA "    ".sprintf("%4s", $counter) . "\n";
    print FA "  </td>\n";
    print FA "  <td width=\"47%\" align=\"left\">\n";
    print FA "    <Font color=grey>(</Font>".sprintf("%.0f", 100*$counter/$length_average)."% <Font color=grey>of the average protein length)</Font>\n";
    print FA "  </td>\n";
    print FA "</tr>\n";
}
print FA "</table>\n";
print FA "<BR>\n";

#3. Calculate columns conservation
print FA "<B>Column conservation statistics:</B><BR>\n";
print FA "<table style=\"width:80%\">\n";
my @cons_threshold = (100, 95, 75, 50);
my %columns_aa_max = ();
foreach my $i (0..$ncolumns-1) {
    my $max_occurence = -1;
    foreach my $aa (keys %{$columns_aa{$i}}) {
        if ($max_occurence == -1 || $max_occurence < $columns_aa{$i}{$aa} ) {
               $max_occurence = $columns_aa{$i}{$aa};
        }
    }
    if ($max_occurence == -1) {
        print "<Font color=red>Error: Failed to calculate conservation index in column ".($i+1)."</Font><BR>\n";
        exit 1;
    }
    $columns_aa_max{$i} = 100*$max_occurence/$nseqs;
}
foreach my $th (@cons_threshold) {
    my $counter = 0;
    foreach my $i (0..$ncolumns-1) {
        if ($columns_aa_max{$i} >= $th) {$counter++;}
    }
    
    print FA "<tr>\n";
    print FA "  <td width=\"47%\" align=\"left\">\n";
    print FA "    <Font color=grey>Number of columns with conservation index at least</Font> ".sprintf("%3s", $th). "% <Font color=grey>:</Font>\n";
    print FA "  </td>\n";
    print FA "  <td width=\"9%\" align=\"center\">\n";
    print FA "    ".sprintf("%4s", $counter)." <Font color=grey>\n";
    print FA "  </td>\n";
    print FA "  <td width=\"44%\" align=\"left\">\n";
    print FA "    (</Font>".sprintf("%.0f", 100*$counter/$length_average)."% <Font color=grey>of the average protein length)</Font><BR>\n";
    print FA "  </td>\n";
    print FA "</tr>\n";
}
print FA "</table>\n";
print FA "<Font color=grey><I>The conservation index for each column is calculated as the occurence of the most frequent amino acid</I></Font><BR>\n";
print FA "<BR>\n";

print "Info: Alignment statistics written to $ANNOTATION\n";
close(FA);

if ($PDB_FOLDER eq "null") {
    if ($VERBOSE) {print "\nDone!\n";}
    exit 0;    
}

print "\n";
print "Info: Creating structure-based annotations of the final alignment\n";
my $HTMLMENU = "$OUTPUT/htmlmenu.html";
my $HTMLMENU_FIRST = "$OUTPUT/htmlmenu_first.php";
open (FM, ">$HTMLMENU") || die "<Font color=red>Error: Failed to open file $HTMLMENU for writing</Font>\n";
print FM "<select class=\"form-control\" id=\"pdbselector\" onchange = loadJSMOL(this.value)>\n";
open (FMF, ">$HTMLMENU_FIRST") || die "<Font color=red>Error: Failed to open file $HTMLMENU_FIRST for writing</Font>\n";
my $first = 0;
my %shannon = ();
foreach my $i (0..$ncolumns-1) {
    if ($columns_gaps{$i} > 5) {
        $shannon{$i} = "null";
        next;
    }
    
    my $curshannon = 0;
    foreach my $aa (keys %{$columns_aa{$i}}) {
        my $p = $columns_aa{$i}{$aa} / $nseqs;
        $curshannon += $p * log(1/$p);
    }
    $shannon{$i} = $curshannon;
}

#Sort pdb names by numeric prefix (\d+)_pdb_chain and everything else by case-insensitive string comparison
my (@nums, @caps);
for (@seqn) {
    push @nums, ( /^(\d+)_/ ? $1 : 999999 );
    push @caps, fc($_);
}
my @sorted = @seqn[ reverse sort {
                   $nums[$b] <=> $nums[$a]
                            ||
                   $caps[$a] cmp $caps[$b]
    } 0..$#seqn
];

foreach my $seqname (@sorted) {    
    my $seq = $aln->{$seqname};
    my $pdb_file = $PDB_FOLDER . "/$seqname.pdb";
        
    if (! -f $pdb_file) {
	if ($VERBOSE) {
	    print "Info: Protein $seqname skipped (reason: no pdb file in $pdb_file)\n";
	}
        next;
    }
    
    my $pdb_file_name = "$seqname.pdb";
    copy ($pdb_file, "$OUTPUT/$pdb_file_name") || die "<Font color=red>Error: Failed to copy file $pdb_file to $OUTPUT/$pdb_file_name</Font>\n";
    
    print "Info: Annotating PDB file for protein $seqname\n";
    
    my %aln2seq_table=();
    my $seq_id = 0;
    foreach my $i (0..$#{$seq}) {
        if ($seq->[$i] eq "-") {
            $aln2seq_table{$i} = -1;
        }
        else {
            $aln2seq_table{$i} = $seq_id;
            $seq_id++;
        }
    }
    
    my $sequence = join("", @{$seq});
    $sequence =~ s/-//g;
    my $seq2pdb_table = alignQueryToPdb($sequence, $pdb_file);
    if (!$seq2pdb_table) {
        print "<Font color=red>Error: Annotation of the PDB file for protein <B>$seqname</B> has been skipped due to errors</Font>\n";
        print "<Font color=red>Error: This error could mean that there is a problem with the format of this structure file downloaded from the PDB.</Font>\n";
        print "<Font color=red>Error: While this problem does not prevent generation of the results (i.e., the alignment), it may cause a local error</Font>\n";
        print "<Font color=red>Error: in the alignment (e.g., missing or rearranged amino acids, etc.). You are advised to download the current results, </Font>\n";
        print "<Font color=red>Error: use this error output to check the corresponding PDB file and, if neccessary, correct the problem, build a new </Font>\n";
        print "<Font color=red>Error: core structural alignment manually, and submit it to the server as a new task in Mode 2.</Font>\n";
        print "\n";
        next;
    }
    
    my $OUTPUT_PY_NAME="$seqname\_annotation.py";
    my $OUTPUT_PY = "$OUTPUT/$OUTPUT_PY_NAME";
    my $OUTPUT_PSE_NAME = "$seqname\_annotation.pse";
    my $OUTPUT_PSE = "$OUTPUT/$OUTPUT_PSE_NAME";
    my $OUTPUT_LOG = "$OUTPUT/$seqname\_annotation.log";
    
    my $c0_x = 255; my $c1_x = 128;
    my $c0_y = 255; my $c1_y = 128;
    my $c0_z = 0;   my $c1_z = 128; #red and light grey in RGB
    
    open(FL, ">$OUTPUT_LOG") or die "<Font color=red>Error: Failed to open file $OUTPUT_LOG for writing</Font>\n";
    print FL "# Basic annotation of PDB $seqname based on the alignment file $FASTA\n";
    print FL "# aln_column - id of the column in the alignment file (first column has aln_column=1)\n";
    print FL "# seq_pos    - id of the position in the protein sequence (first column has seq_pos=1)\n";
    print FL "# pdb_pos    - id of the position in the protein structure $pdb_file_name\n";
    print FL "# %_of_gaps  - frequency of gaps in the column. Data shown only for alignment columns with no more than 5% of gaps.\n";
    print FL "# %_of_cons  - conservation index of a column calculated as the occurence of the most frequent amino acid\n";
    print FL "# H          - Shannon entropy value, a basic measure of sequence variability\n";
    print FL "#              (0 means no variability=totally conserved, ".sprintf("%.2f", $SHANNON_MAX)." means all 20 amino acids are present in equall proportions)\n";
    print FL "#\n";
    print FL sprintf("# %10s %10s %15s %10s %10s %10s\n", "aln_column", "seq_pos", "pdb_pos", "%_of_gaps", "%_of_cons", "H");        
    
    open(FP, ">$OUTPUT_PY") or die "<Font color=red>Error: Failed to open file $OUTPUT_PY for writing</Font>\n";
    print FP "cmd.load(\"$pdb_file_name\", \"protein\")\n";
    print FP "cmd.color(\"red\", \"protein and name CA\")\n";
    print FP "cmd.show(\"cartoon\")\n";
    print FP "cmd.hide(\"lines\")\n";
    print FP "cmd.select(\"HETATOMS\", \"hetatm\")\n";
    print FP "cmd.color(\"chartreuse\", \"HETATOMS and element C\")\n";
    print FP "cmd.color(\"red\", \"HETATOMS and element O\")\n";
    print FP "cmd.color(\"blue\", \"HETATOMS and element N\")\n";
    print FP "cmd.show(\"sticks\", \"HETATOMS\")\n";
    
    my $firstConserved=0;
    foreach my $i (0 .. $ncolumns-1) {
        
        if ($shannon{$i} eq "null") {
            #Column $i contains more than 5% of gaps
            next;
        }
        
        my $aln_id = $i;            
        my $seq_id = $aln2seq_table{$aln_id};
        
        if ($seq_id eq "-1") {
            print FL sprintf("  %10s %10s %15s %10s %10s %10s\n", ($aln_id+1), "gap", "gap", sprintf("%.2f", $columns_gaps{$aln_id}), sprintf("%.2f", $columns_aa_max{$i}), sprintf("%.2f", $shannon{$aln_id}));
            next;
        }
        
        my $pdb_command = $seq2pdb_table->{$seq_id};
        
        if ($pdb_command eq "null") {
            print FL sprintf("  %10s %10s %15s %10s %10s %10s\n", ($aln_id+1), ($seq_id+1), "gap", sprintf("%.2f", $columns_gaps{$aln_id}), sprintf("%.2f", $columns_aa_max{$i}), sprintf("%.2f", $shannon{$aln_id}));            
            next;
        }
        
        print FL sprintf("  %10s %10s %15s %10s %10s %10s\n", ($aln_id+1), ($seq_id+1), $pdb_command, sprintf("%.2f", $columns_gaps{$aln_id}), sprintf("%.2f", $columns_aa_max{$i}), sprintf("%.2f", $shannon{$aln_id}));            
        
        my $coeff;
        if ($shannon{$i} >= $SHANNON_MAX/2) {$coeff=0;}
        else {
            $coeff=($shannon{$i} -  $SHANNON_MAX/2)/($SHANNON_MIN - $SHANNON_MAX/2);
        }                    
        my $ci_x = $c1_x + $coeff*($c0_x - $c1_x);
        my $ci_y = $c1_y + $coeff*($c0_y - $c1_y);
        my $ci_z = $c1_z + $coeff*($c0_z - $c1_z);
    
        print FP "cmd.set_color(\"temp_color$i\", \"[$ci_x,$ci_y,$ci_z]\")\n";    
        print FP "cmd.color(\"temp_color$i\", \"protein and $pdb_command and element C\")\n";
        if ($columns_aa_max{$i} >= 95) {
    	    print FP "cmd.show(\"sticks\", \"protein and $pdb_command\")\n";
    	    if ($firstConserved==0) {
    		print FP "cmd.select(\"conserved95\", \"protein and $pdb_command\")\n";
    		$firstConserved=1;
    	    }
    	    else {
    		print FP "cmd.select(\"conserved95\", \"conserved95 or (protein and $pdb_command)\")\n";
    	    }
    	}
    }
    print FP "cmd.hide(\"nonbonded\")\n";
    print FP "cmd.hide(\"lines\")\n";
    print FP "cmd.center(\"all\")\n";
    print FP "cmd.select(\"none\")\n";
    print FP "cmd.do(\"bg_color white\")\n";
    print FP "cmd.save(\"$OUTPUT_PSE_NAME\")\n";
    
    close(FP);
    close(FL);
    
    if ( ! -e $OUTPUT_PY || -z $OUTPUT_PY ){
        print "<Font color=red>Error: Failed to create Pymol script $OUTPUT_PY_NAME</Font>\n";
        exit 1;
    }
    
    #Create Pymol session
    system("cd $OUTPUT; pymol -cq $OUTPUT_PY_NAME > /dev/null;");
    
    if ( ! -e $OUTPUT_PSE || -z $OUTPUT_PSE ){
        print "<Font color=red>Warning: Failed to create Pymol session $OUTPUT_PSE_NAME</Font>\n";
    }
    else {
	#Add new option to the web-menu
	print FM "<option value=\"$seqname\">$seqname</option>\n";
	if (!$first) {
	    $first=1;
	    print FMF "$seqname";
	}
    }
}
print FM "<select>\n";
print FM "<small class=\"form-text text-muted\">Choose a protein from the dropdown menu and allow some time for loading of the content</small>\n";
close(FM);
close(FMF);

#===================================================================
# SUBROUTINES
#===================================================================

#Returns a hash table
#key = aa id in the sequence starting from 0 (ZERO!!!)
#function = pdb identifier
sub alignQueryToPdb {
    #Retreive the first and second arguments passed to the method and store them in separate variables
    my $sequence = shift;
    my $pdb_file = shift;
    
    #Execute the subroutine to read in the pdb file
    my $pdb = readpdb($pdb_file);
    
    my %residues = ();
    my %check = ();
    my $rank = 0;
    foreach my $id (sort {$a <=> $b} keys %{$pdb}) {
        my $atom = $pdb->{$id}->{"atom"};
        if ($pdb->{$id}->{"atom"} ne "CA" || $pdb->{$id}->{"head"} ne "ATOM") {next;} #Skip if this record does not correspond to a CA atom
                
        my $resname = $pdb->{$id}->{"res"};
        my $resid = $pdb->{$id}->{"resn"};
                
        #Check for residues with duplicate resid and remove if it reffers to alternative conformation
        if ($check{$resid} && $resname ne $check{$resid}) {            
            if ($check{$resid} =~ /^[A-Z](...)/) {
                my $name1 = $1;
                if ($resname =~ /^[A-Z](...)/) {
                    my $name2 = $1;
                    if ($name1 eq $name2) {
                        if ($VERBOSE) {print "<Font color=red>Warning: Omitting duplicate CA atom entry for residue #$resid ($check{$resid} and $resname)</Font>\n";}
                        next;
                    }
                }
            }
        }
                        
        #Store resid and resname to the hash
        #Hashkey = rank of the residue in the pdb file
        $residues{$rank}{0} = $resid;
        $residues{$rank}{1} = $resname;
        
        $check{$resid} = $resname;
        
        $rank++; #Increment the id for the next iteration        
    }
    
    #Now parse to check three-letter codes
    my %oneletter = (
        GLY => 'G',
        ALA => 'A',
        VAL => 'V',
        LEU => 'L',
        ILE => 'I',
        PHE => 'F',
        PRO => 'P',
        SER => 'S',
        THR => 'T',
        TYR => 'Y',
        CYS => 'C',
        MET => 'M',
        LYS => 'K',
        ARG => 'R',
        HIS => 'H',
        TRP => 'W',
        ASP => 'D',
        GLU => 'E',
        ASN => 'N',
        GLN => 'Q');    
    
    my $pdb_sequence = "";
    foreach my $id (sort {$a <=> $b} keys %residues) {
        my $resname = $residues{$id}{1};
        
        #If the residue name is three letters - is OK, just skip
        if ($resname =~ /^...$/) {            
        }
        
        #If the residue name is four letters - remove the first one (left from alternative conformation)
        elsif ($resname =~ /^[A-Z](...)$/) {
            $residues{$id}{1} = $1;
        }
        
        else {
            print "<Font color=red>Error: Invalid residue code $resname for resid $residues{$id}{0}</Font>\n";
            return;
        }
        
        $resname = $residues{$id}{1};
        if (!$oneletter{$resname}) {
            print "<Font color=red>Error: Unknown residue code $resname</Font>\n";
            return;
        }
        
        $pdb_sequence .= $oneletter{$resname};
    }
        
    if ($VERBOSE) {print "Info: Total number of residues in PDB file is $rank\n";}
    if ($VERBOSE) {print "Info: PDB sequence is   $pdb_sequence\n";}
    
    #Now align the two sequences
    open(FTEMP, ">$MASK\_temp.seq") or die "<Font color=red>Error: Failed to open file $MASK\_temp.seq for writing</Font>\n";
    print FTEMP ">in_alignment\n$sequence\n>in_pdb\n$pdb_sequence\n";
    close(FTEMP);
    system("$MAFFT --auto --quiet $MASK\_temp.seq > $MASK\_temp.fasta 2> /dev/null");
    my $aln = readaln("$MASK\_temp.fasta");    
    my $query_aligned = $aln->{"in_alignment"};
    my $pdb_aligned = $aln->{"in_pdb"};
    
    my $mismatch_error = 0;
    my @aln_print = ();
    my @pdb_print = ();
    
    my %reference = ();    
    my $i = 0;
    my $j = 0;
    my @query_split = split(//, $sequence);
    foreach my $col (0..$#$query_aligned) {        
        if ($query_aligned->[$col] eq "-" && $pdb_aligned->[$col] eq "-") {            
            print "<Font color=red>Internal Error: Alignment is invalid in column ".($col+1)." as it contains only gaps</Font>\n";
            system("cat $MASK\_temp.fasta");
            return;
        }
        elsif ($query_aligned->[$col] eq "-") {
            if ($VERBOSE) {print "<Font color=red>Warning: PDB residue ". $residues{$j}{1}."`".$residues{$j}{0}. " does not have a pair in the Query sequence</Font>\n";}
            
            push (@aln_print, $query_aligned->[$col]);
            push (@pdb_print, $pdb_aligned->[$col]);

            $j++;
        }
        elsif ($pdb_aligned->[$col] eq "-") {
            if ($VERBOSE) {print "<Font color=red>Warning: Query residue ". $query_split[$i].$i. " does not have a pair in the PDB sequence</Font>\n";}
            $reference{$i} = "null";
            
            push (@aln_print, $query_aligned->[$col]);
            push (@pdb_print, $pdb_aligned->[$col]);
            
            $i++;
        }
        elsif ($query_aligned->[$col] eq $pdb_aligned->[$col]) {
            if ($query_split[$i] ne $query_aligned->[$col]) {
                print "<Font color=red>Internal error: query sequences are incompatible in column ".($col+1)."</Font>\n";
                system("cat $MASK\_temp.fasta");
                return;
            }
            if ($query_split[$i] ne $oneletter{$residues{$j}{1}}) {
                print "<Font color=red>Internal error: query and pdb sequences are incompatible in column ".($col+1)."</Font>\n";
                system("cat $MASK\_temp.fasta");
                return;
            }
            
            #print "$col $query_aligned->[$col] $query_split[$i]$i $oneletter{$residues{$j}{1}} "."///".$residues{$j}{1}."`".$residues{$j}{0}."\n";
            
            push (@aln_print, "<Font color=green>".$query_aligned->[$col]."</Font>");
            push (@pdb_print, "<Font color=green>".$pdb_aligned->[$col]."</Font>");
            
            $reference{$i} = "////".$residues{$j}{1}."`".$residues{$j}{0};            
            $i++;
            $j++;
        }
        elsif ($query_aligned->[$col] ne $pdb_aligned->[$col]) {
            
            push (@aln_print, "<Font color=red><B>".$query_aligned->[$col]."</B></Font>");
            push (@pdb_print, "<Font color=red><B>".$pdb_aligned->[$col]."</B></Font>");

            print "<Font color=red>Error: The query (amino acid $query_aligned->[$col]) and pdb (amino acid $pdb_aligned->[$col], $residues{$j}{1}`$residues{$j}{0}) sequences mismatch at column ". ($col+1) ." of the pairwise alignment</Font>\n";
	    if ($query_aligned->[$col] eq "X") {
		print "<Font color=red>Error: The \"X\" mismatch error is usually caused by unknown residues or ligands containing a \"CA\" atom in the PDB.</Font>\n";
	    }
            $mismatch_error = 1;
	    
	    $i++;
            $j++;
        }        
    }

    if ($i != $#query_split+1) {
        print "<Font color=red>Internal error: query sequence length mismatch (".$i." ". ($#query_split+1).")</Font>\n";
        return;
    }

    if ($mismatch_error) {
    
	my @aln_print_string = ();
	my @pdb_print_string = ();
    
	if (scalar @aln_print != scalar @pdb_print) {
    	    print "<Font color=red>Internal error: Print arrays (level 1) have different length</Font>\n";
    	    return;
	}
    
	my $s = 0;
	my $string1 = "";
	my $string2 = "";
	foreach my $k (0..$#aln_print) {
        
    	    $string1 .= $aln_print[$k];
    	    $string2 .= $pdb_print[$k];
        
    	    $s++;
    	    if ($s == 60)
    	    {            
        	push(@aln_print_string, $string1);
        	push(@pdb_print_string, $string2);
        	$string1 = "";
        	$string2 = "";
        	$s = 0;
    	    }
	}
    
	if ($string1 ne "" || $string2 ne "")
	{            
    	    push(@aln_print_string, $string1);
    	    push(@pdb_print_string, $string2);
    	    $string1 = "";
    	    $string2 = "";
	}
    
	if (scalar @aln_print_string != scalar @pdb_print_string) {
    	    print "<Font color=red>Internal error: Print arrays (level 2) have different length</Font>\n";
    	    return;
	}
    
	my $space = "&nbsp;";
	#print "<Font face=\"Courier\">\n";
	foreach my $k (0..$#aln_print_string) {
    	    print ">best_match:$space$aln_print_string[$k]\n";
    	    print ">pdb$space$space$space$space$space$space$space:$space$pdb_print_string[$k]\n";
    	    print "\n";
	}
	#print "</Font>\n";
    
	return;
    }
        
    
    if ($VERBOSE) {print "Info: Query sequence and PDB sequence have been successfully aligned\n";}
    
    #Delete temporary files
    unlink("./$MASK\_temp.seq");
    unlink("./$MASK\_temp.fasta");    
    
    return \%reference;
}

#Read in a pdb file
sub readpdb {
    my $input = shift;
    
    my %pdb = ();
    open(FI, $input) or die "<Font color=red>Error: Can not open file $input for reading</Font>\n\n";
    
    #Read in the PDB file
    
    while(<FI>)
    {
        chomp;                
        
        #Remove Windows ^M sign
        s/\r//g;
        
        my $keyword = substr($_,0,6);    # look for ATOM record
        if ($keyword eq "ATOM  " || $keyword eq "HETATM") {
        
            my $length = length $_;
            if ($length < 54) {                
                print "\n$_\n";
                print "<Font color=red>Error: ATOM and HETATM fields must be at least 54 characters in length (found $length))</Font>\n";
                pdbFormatError();
            }
            
            my $num = substr($_,6,5);
            $num =~ s/\s+//g;
            while ($pdb{$num}) {
                $num++;
            }

            $pdb{$num}{"head"} = substr($_,0,6);
            $pdb{$num}{"atom"} = substr($_,12,4);
            my $type = $pdb{$num}{"atom"};
            $type=substr($type,0,1);
            $pdb{$num}{"type"} = $type;
            $pdb{$num}{"res"} = substr($_,16,4);
            $pdb{$num}{"chainid"} = substr($_,21,1);
            $pdb{$num}{"resn"} = substr($_,22,4);
            $pdb{$num}{"xcor"} = substr($_,30,8);
            $pdb{$num}{"ycor"} = substr($_,38,8);
            $pdb{$num}{"zcor"} = substr($_,46,8);
            #$pdb{$num}{"occ"} = substr($_,54,6);
            #$pdb{$num}{"bfac"} = substr($_,60,6);
            #$pdb{$num}{"segid"} = substr($_,72,4);
            #$pdb{$num}{"elm"} = substr($_,76,2);
            #$pdb{$num}{"charge"} = substr($_,78,2);
            
            #Remove spaces
            foreach my $key (keys %{$pdb{$num}}) {
                $pdb{$num}{$key} =~ s/\s+//g;
            }
        }
    }
    close(FI);
    
    return \%pdb;
}

sub pdbFormatError{    
    print "<Font color=red>Error: The file submitted as PDB is not valid</Font>\n\n";    
    exit 1;
}

sub readaln
{
    my $filename = shift;
    open(FI, $filename) or die "<Font color=red>Error: Can not open file $filename for reading</Font>\n\n";;
    my %hash = ();
    my %hash_check=();
    my $name = "";
    my $seq = "";

    while(<FI>)
    {
        chomp;

        #Remove Windows ^M sign
        s/\r//g;

        if (/^>(.+)/) 
        {
            if (!$name) {$name = $1; next;}
            my $tmp = $1;            
                        
            my @arr=split(//,$seq);
            #if ($hash{$name}) {push (@{$hash{$name}}, $chain);} # Insert chain delemeter after the first chain
            push (@{$hash{$name}}, @arr); # _Append_ new lines to the hash
            
            if ($hash_check{$name}) {
                print "<Font color=red>Error: Sequence name \"$name\" used twice in one alignment</Font>\n";
                fastaFormatError();            
            }
            
            $hash_check{$name}=1;
            @arr=();
            $name = $tmp;
            $seq = "";
            next;
        }    
        #if (!$name) {die "Hard!!! Sequence without a name:\n$_\n";}
                
        #Convert to upper case
        $seq.= uc $_;    
    }
    if ($name) {
        my @arr=split(//,$seq);
        push (@{$hash{$name}}, @arr); # _Append_ new lines to the hash
        if ($hash_check{$name}) {
            print "<Font color=red>Error: Sequence name \"$name\" used twice in one alignment</Font>\n";
            fastaFormatError();            
        }
    }
    else {
        print "<Font color=red>Error: Unexpected end-of-file</Font>\n";
        fastaFormatError();            
    }
    
    return \%hash;
}

#Get the sequence names in order they appear in the alignment file
sub getSeqOrder
{
    my $filename = shift;
    open(FI, $filename);
    my @names = ();
    my %hash_check=();
    my $name = "";

    while(<FI>)
    {
        chomp;

        #Remove Windows ^M sign
        s/\r//g;    

        if (/^>(.+)/) 
        {
            if (!$name) {$name = $1; next;}
            my $tmp = $1;
            
            push (@names, $name); # _Append_ new lines to the hash
            
            if ($hash_check{$name}) {                
                print "<Font color=red>Error: Sequence name \"$name\" is used twice in one alignment</Font>\n";
                fastaFormatError();
            }
            
            $hash_check{$name}=1;            
            $name = $tmp;            
            next;
        }    
        if (!$name) {
            print "<Font color=red>Error: Sequence without a name:\n$_</Font>\n";
            fastaFormatError();
        }
    }
    
    push (@names, $name); # _Append_ new lines to the hash
    
    return \@names;
}

sub fastaFormatError{

    print "<Font color=red>Error: The file submitted is not a valid FASTA Alignment </Font>\n";
    print "<Font color=red>Error: You can read more about the <a href=\"http://en.wikipedia.org/wiki/FASTA_format\">FASTA format</a> and <a href=\"http://en.wikipedia.org/wiki/Multiple_sequence_alignment\">Multiple sequence alignments</a></Font>\n";
    print "<Font color=red>Error: Please revise your data and resubmit in a new task</Font>\n\n";
    
    exit 1;
}